import os
import pathlib
from typing import Any, Generator, List, Optional, Tuple
from unittest import mock

import pytest
from click.testing import CliRunner

from gcs_migration.commands.gcs_migration import main
from gcs_migration.models import MigrationPlan


@pytest.fixture
def mock_map_identities_environment(
    mock_get_transfer_api: Any,
    mock_get_auth_api: Any,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    tmp_path: pathlib.Path,
) -> Generator[MigrationPlan, None, None]:
    with mock.patch.object(MigrationPlan, "load") as mock_migration_load:
        mock_migration_load.return_value = migration_plan_cmd_create
        migration_plan_cmd_create.root = tmp_path
        os.mkdir(tmp_path / "migration_plan")
        yield migration_plan_cmd_create


@pytest.mark.parametrize(
    "subject,globus_username,local_username,expect",
    [
        # Success: mapping exists, globus_username lookup succeeds
        ("/C=US/O=Globus/OU=Test/CN=User1", "user@globus.org", "local1", 0),
        # Failure: no mapping exists (subject does not map to local_username)
        ("/C=US/O=Globus/OU=Test/CN=User1", "user@globus.org", "badlocal1", 1),
        # Failure: mapping exists, globus_username lookup fails
        ("/C=US/O=Globus/OU=Test/CN=User1", "unknownuser@globus.org", "local1", 1),
        # Failure: no mapping exists (no such subject)
        ("/C=US/O=Globus/OU=Test/CN=Unknown User1", "user@globus.org", "local1", 1),
        # Success: mapping exists (multiple), globus_username lookup succeeds
        ("/C=US/O=Globus/OU=Test/CN=User2", "user@globus.org", "local2", 0),
        # Success: mapping exists, globus_username lookup succeeds, but email domain
        # not added to allowed_domains
        ("/C=US/O=Globus/OU=Test/CN=Email1", "user@email.org", "local9", 0),
    ],
)
def test_identity_mapping_gridmap_full(
    mock_map_identities_environment: MigrationPlan,
    subject: str,
    globus_username: str,
    local_username: str,
    expect: int,
) -> None:
    migration_plan = mock_map_identities_environment
    domain = globus_username.rsplit("@", 1)[1]
    assert migration_plan.storage_gateway is not None

    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "gridmap"]

    assert migration_plan.gsi_authz_configuration
    assert migration_plan.gsi_authz_configuration.gridmap_configuration
    assert migration_plan.gsi_authz_configuration.gridmap_configuration[0]

    if expect == 0:
        assert migration_plan.guest_collections is not None
        assert len(migration_plan.guest_collections) > 0
        assert migration_plan.guest_collections[0].credential is not None
        migration_plan.guest_collections[0].credential.gsi_authz_callout_index = 1

    args.extend(["-f", subject, globus_username, local_username])

    result = runner.invoke(main, args, catch_exceptions=False)
    assert result.exit_code == expect

    assert migration_plan.storage_gateway.allowed_domains is not None

    found_mapping = False
    for m in migration_plan.storage_gateway.identity_mappings:
        if (
            m["mappings"][0]["match"] == globus_username
            and m["mappings"][0]["output"] == local_username
        ):
            found_mapping = True

    if expect == 0:
        assert found_mapping
        if domain == "email.org":
            assert domain not in migration_plan.storage_gateway.allowed_domains
        else:
            assert domain in migration_plan.storage_gateway.allowed_domains
        if local_username == "local1":
            assert migration_plan.guest_collections is not None
            assert len(migration_plan.guest_collections) > 0
            assert (
                migration_plan.guest_collections[0].credential.identity_username
                == globus_username
            )
    else:
        assert not found_mapping
        assert domain not in migration_plan.storage_gateway.allowed_domains


@pytest.mark.parametrize(
    "subject,globus_username,local_username,expect",
    [
        # Success: mapping exists, globus_username lookup succeeds
        ("/C=US/O=Globus/OU=Test/CN=User1", "user@globus.org", "local1", 0),
        # Failure: mapping exists, globus_username lookup fails
        ("/C=US/O=Globus/OU=Test/CN=User1", "unknownuser@globus.org", "local1", 1),
        # Failure: no mapping exists (no such subject)
        ("/C=US/O=Globus/OU=Test/CN=Unknown User1", "user@globus.org", "local1", 1),
        # Failure: mapping exists (multiple), but not to 'local1'
        ("/C=US/O=Globus/OU=Test/CN=User2", "user@globus.org", "local1", 1),
        # Success: mapping exists, globus_username lookup succeeds, but email domain
        # not added to allowed_domains
        ("/C=US/O=Globus/OU=Test/CN=Email1", "user@email.org", "local9", 0),
    ],
)
def test_identity_mapping_gridmap_subject(
    mock_map_identities_environment: MigrationPlan,
    subject: str,
    globus_username: str,
    local_username: str,
    expect: int,
) -> None:
    migration_plan = mock_map_identities_environment
    domain = globus_username.rsplit("@", 1)[1]
    assert migration_plan.storage_gateway is not None

    migration_plan.storage_gateway.allowed_domains = None
    migration_plan.storage_gateway.identity_mappings = None

    if expect == 0:
        assert migration_plan.guest_collections is not None
        assert len(migration_plan.guest_collections) > 0
        assert migration_plan.guest_collections[0].credential is not None
        migration_plan.guest_collections[0].credential.gsi_authz_callout_index = 1

    runner = CliRunner()
    args: List[str] = ["identity-mapping", "gridmap"]

    assert migration_plan.gsi_authz_configuration
    assert migration_plan.gsi_authz_configuration.gridmap_configuration
    assert migration_plan.gsi_authz_configuration.gridmap_configuration[0]

    args.extend(["-s", subject, globus_username])

    result = runner.invoke(main, args, input="y\n")
    assert result.exit_code == expect

    assert migration_plan.storage_gateway.allowed_domains is not None

    found_mapping = False
    for m in migration_plan.storage_gateway.identity_mappings:
        if (
            m["mappings"][0]["match"] == globus_username
            and m["mappings"][0]["output"] == local_username
        ):
            found_mapping = True

    if expect == 0:
        assert found_mapping
        if domain == "email.org":
            assert domain not in migration_plan.storage_gateway.allowed_domains
        else:
            assert domain in migration_plan.storage_gateway.allowed_domains
        if local_username == "local1":
            assert migration_plan.guest_collections is not None
            assert len(migration_plan.guest_collections) > 0
            assert (
                migration_plan.guest_collections[0].credential.identity_username
                == globus_username
            )
    else:
        assert not found_mapping
        assert domain not in migration_plan.storage_gateway.allowed_domains


def test_requires_logged_in(mock_not_logged_in: None) -> None:
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "identity-mapping",
            "gridmap",
            "-s",
            "/C=US/O=Globus/OU=Test/CN=User1",
            "user@globus.org",
        ],
        catch_exceptions=False,
    )
    assert "you will have to login" in result.output
