import os
import pathlib

from typing import Callable, List, Union
from typing_extensions import Protocol

import pytest

from gcs_migration.parsers import CalloutDict


class CalloutLine(Protocol):
    config: CalloutDict

    def __str__(self) -> str:
        ...


@pytest.fixture
def cilogon_callout(
    incommon_urlopener: None, test_host_cert_file: Union[str, os.PathLike]
) -> CalloutLine:
    c: CalloutDict = {
        "callout_type": "globus_mapping",
        "mandatory": False,
        "library": "libglobus_gridmap_eppn_callout",
        "symbol": "globus_gridmap_eppn_callout",
        "environment": {
            "GLOBUS_MYPROXY_CA_CERT": str(test_host_cert_file),
            "GLOBUS_MYPROXY_AUTHORIZED_DN": "/DC=org/DC=cilogon/C=US/O=Ohio State University",
        },
        "raw": "",
    }
    env_string = " ".join([f'{k}="{v}"' for k, v in c["environment"].items()])
    raw = f"{c['callout_type']} {c['library']} {c['symbol']} ENV:{env_string}"
    c["raw"] = raw

    class CILogonCalloutLine:
        config = c

        def __str__(self) -> str:
            return c["raw"]

    return CILogonCalloutLine()


@pytest.fixture
def myproxy_callout(test_host_cert_file: str) -> CalloutLine:
    c: CalloutDict = {
        "callout_type": "globus_mapping",
        "mandatory": False,
        "library": "libglobus_gridmap_verify_myproxy_callout",
        "symbol": "globus_gridmap_verify_myproxy_callout",
        "environment": {"GLOBUS_MYPROXY_CA_CERT": str(test_host_cert_file)},
        "raw": "",
    }
    env_string = " ".join([f'{k}="{v}"' for k, v in c["environment"].items()])

    raw = f"{c['callout_type']} {c['library']} {c['symbol']} ENV:{env_string}"

    c["raw"] = raw

    class MyProxyCalloutLine:
        config = c

        def __str__(self) -> str:
            return c["raw"]

    return MyProxyCalloutLine()


@pytest.fixture
def myproxy_callout_without_env(test_host_cert_file: str) -> CalloutLine:
    c: CalloutDict = {
        "callout_type": "globus_mapping",
        "mandatory": False,
        "library": "libglobus_gridmap_verify_myproxy_callout",
        "symbol": "globus_gridmap_verify_myproxy_callout",
        "environment": {},
        "raw": "",
    }

    raw = f"{c['callout_type']} {c['library']} {c['symbol']}"

    c["raw"] = raw

    class MyProxyCalloutLine:
        config = c

        def __str__(self) -> str:
            return c["raw"]

    return MyProxyCalloutLine()


@pytest.fixture
def unknown_callout(test_host_cert_file: str) -> CalloutLine:
    c: CalloutDict = {
        "callout_type": "globus_mapping",
        "mandatory": False,
        "library": "libglobus_unsupported_callout",
        "symbol": "globus_unsupported_callout",
        "environment": {},
        "raw": "",
    }

    raw = f"{c['callout_type']} {c['library']} {c['symbol']}"

    c["raw"] = raw

    class UnknownCalloutLine:
        config = c

        def __str__(self) -> str:
            return c["raw"]

    return UnknownCalloutLine()


GsiAuthzConfFixtureCreator = Callable[[List[CalloutLine]], pathlib.Path]


@pytest.fixture
def gsi_authz_conf(tmp_path: pathlib.Path) -> GsiAuthzConfFixtureCreator:
    def _gsi_authz_conf(config_lines: List[CalloutLine]) -> pathlib.Path:
        name = tmp_path / "gsi_authz.conf"

        with open(name, "w") as f:
            for line in config_lines:
                print(str(line), file=f)
        return name

    return _gsi_authz_conf


@pytest.fixture
def myproxy_gsi_authz_conf(
    myproxy_callout: CalloutLine, gsi_authz_conf: GsiAuthzConfFixtureCreator
) -> pathlib.Path:
    return gsi_authz_conf([myproxy_callout])
